"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const express_1 = __importDefault(require("express"));
const express_history_api_fallback_1 = __importDefault(require("express-history-api-fallback"));
const get_port_1 = __importDefault(require("get-port"));
const http_proxy_1 = __importDefault(require("http-proxy"));
const path_1 = require("path");
const webpack_1 = __importDefault(require("webpack"));
const webpack_dev_server_1 = __importDefault(require("webpack-dev-server"));
const ui_aspect_1 = require("./ui.aspect");
const webpack_dev_config_1 = require("./webpack/webpack.dev.config");
class UIServer {
    constructor(graphql, expressExtension, ui, uiRoot, uiRootExtension, logger) {
        this.graphql = graphql;
        this.expressExtension = expressExtension;
        this.ui = ui;
        this.uiRoot = uiRoot;
        this.uiRootExtension = uiRootExtension;
        this.logger = logger;
        this._port = 0;
    }
    get port() {
        return this._port;
    }
    getDevConfig() {
        return __awaiter(this, void 0, void 0, function* () {
            const aspects = yield this.uiRoot.resolveAspects(ui_aspect_1.UIRuntime.name);
            const aspectsPaths = aspects.map((aspect) => aspect.aspectPath);
            return webpack_dev_config_1.devConfig(this.uiRoot.path, [yield this.ui.generateRoot(aspects, this.uiRootExtension)], this.uiRoot.name, aspectsPaths);
        });
    }
    start({ port } = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            const app = this.expressExtension.createApp();
            const selectedPort = yield this.selectPort(port || 4000);
            const root = path_1.join(this.uiRoot.path, '/public');
            const server = yield this.graphql.createServer({ app });
            yield this.configureProxy(app, server);
            app.use(express_1.default.static(root));
            app.use(express_history_api_fallback_1.default('index.html', { root }));
            server.listen(selectedPort);
            this._port = selectedPort;
            this.logger.info(`UI server of ${this.uiRootExtension} is listening to port ${selectedPort}`);
        });
    }
    configureProxy(app, server) {
        return __awaiter(this, void 0, void 0, function* () {
            const proxServer = http_proxy_1.default.createProxyServer();
            proxServer.on('error', (e) => this.logger.error(e.message));
            const proxyEntries = this.uiRoot.getProxy ? yield this.uiRoot.getProxy() : [];
            server.on('upgrade', function (req, socket, head) {
                const entry = proxyEntries.find((proxy) => proxy.context.some((item) => item === req.url));
                if (!entry)
                    return;
                proxServer.ws(req, socket, head, {
                    target: entry.target,
                });
            });
            proxyEntries.forEach((entry) => {
                entry.context.forEach((route) => {
                    app.use(`${route}/*`, (req, res) => {
                        proxServer.web(req, res, Object.assign(Object.assign({}, entry), { target: `${entry.target}/${req.originalUrl}` }));
                    });
                });
            });
        });
    }
    dev({ port } = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            const selectedPort = yield this.selectPort(port);
            yield this.start({ port: yield get_port_1.default({ port: 4000 }) });
            const config = yield this.getDevConfig();
            const compiler = webpack_1.default(config);
            const devServerConfig = yield this.getDevServerConfig(config.devServer);
            const devServer = new webpack_dev_server_1.default(compiler, devServerConfig);
            devServer.listen(selectedPort);
            return devServer;
        });
    }
    selectPort(port) {
        return __awaiter(this, void 0, void 0, function* () {
            if (port)
                return port;
            return get_port_1.default({ port: get_port_1.default.makeRange(3000, 3200) });
        });
    }
    getProxy() {
        var _a, _b;
        return __awaiter(this, void 0, void 0, function* () {
            const proxyEntries = (yield ((_b = (_a = this.uiRoot).getProxy) === null || _b === void 0 ? void 0 : _b.call(_a))) || [];
            const gqlProxies = [
                {
                    context: ['/graphql', '/api'],
                    target: 'http://localhost:4000',
                    changeOrigin: true,
                },
                {
                    context: ['/subscriptions'],
                    target: 'ws://localhost:4000',
                    ws: true,
                },
            ];
            return gqlProxies.concat(proxyEntries);
        });
    }
    getDevServerConfig(config) {
        return __awaiter(this, void 0, void 0, function* () {
            const proxy = yield this.getProxy();
            const devServerConf = Object.assign(Object.assign({}, config), { proxy });
            return devServerConf;
        });
    }
    static create(props) {
        return new UIServer(props.graphql, props.express, props.ui, props.uiRoot, props.uiRootExtension, props.logger);
    }
}
exports.UIServer = UIServer;
//# sourceMappingURL=ui-server.js.map